/* jshint asi: true */
/* jshint esversion: 6 */
/* jshint node: true */
/* jshint jquery: true */
/* jshint strict: false */
/* jshint eqeqeq: false */
//=====================================================================================================
// Electron
//=====================================================================================================
const fs = require('fs')
const electron = require('electron')

const remote = electron.remote;
const Menu = remote.Menu
const MenuItem = remote.MenuItem
const clipboard = remote.clipboard
//=====================================================================================================
// Query Selector Shorthands
//=====================================================================================================
Node.prototype.q =function (selector) {
	 return this.querySelector(selector);
}
Node.prototype.qa = function (selector) {
	return this.querySelectorAll(selector);
}

String.prototype.GetCoreClass = function(){
	let re = /(AssetObjects[\W]+)/g
	return this.replace(re, "")
}

String.prototype.CleanClasses = function(){
	let re = /(AssetObjects)([\W]+)(\w+)/g
	return this.replace(re, "$1..$3")
}

// String.prototype.CleanTrailing = function(){
// 	let re = /(<\/AssetObjects\.\.ArtDefSet>)([\w\d\s]+)/g
// 	return this.replace(re, "$1")
// }
//=====================================================================================================
// Color Functions
//=====================================================================================================
function componentToHex(c) {
    var hex = c.toString(16);
    return hex.length == 1 ? "0" + hex : hex;
}

function rgbToHex(r, g, b) {
    return "#" + componentToHex(r) + componentToHex(g) + componentToHex(b);
}

function hexToRgb(hex) {
	// Expand shorthand form (e.g. "03F") to full form (e.g. "0033FF")
	var shorthandRegex = /^#?([a-f\d])([a-f\d])([a-f\d])$/i;
	hex = hex.replace(shorthandRegex, function(m, r, g, b) {
		return r + r + g + g + b + b;
	});

	var result = /^#?([a-f\d]{2})([a-f\d]{2})([a-f\d]{2})$/i.exec(hex);
	return result ? {
		r: parseInt(result[1], 16),
		g: parseInt(result[2], 16),
		b: parseInt(result[3], 16)
	} : null;
}
//=====================================================================================================
// Sidebar
//=====================================================================================================
customElements.define('sidebarlist-li',
class extends HTMLElement {
	constructor() {
		super();
	}

	def(sName, pXML, bIsRoot) {
		var template = document.getElementById('sidebarlist_li').content
		this.appendChild(template.cloneNode(true))

		// Name //
		this.Name = sName;
			this.q(".li_wrapper > .li_label").innerHTML = this.Name
		// Node //
		this.XML_Node = pXML
		// Is Root //
		if (bIsRoot){this.classList.add("RootCollection")}
		// On Click
		var pLI = this
		var li_wrapper_OnClick = this.li_wrapper_OnClick
		this.q(".li_wrapper").onclick = function() {
			li_wrapper_OnClick(pLI)
		}

		return this
	}	

	//-----------------------------------------------------------------------
	//-----------------------------------------------------------------------

	li_wrapper_OnClick(pLI) {

		var pWrapper = pLI.q(".li_wrapper")

		if (pWrapper.parentElement.classList.contains("li_selected")){
			for (let pChild of pWrapper.qa(".li_selected").entries()){
				pChild[1].classList.remove("li_selected")
			}

			pWrapper.parentElement.classList.remove("li_selected")
			return	
		}

		var pList = document.getElementById("sidebarlist")

		// Hide all uls
		for (let pChild of pList.qa(".li_selected").entries()){
			pChild[1].classList.remove("li_selected")
		}
		for (let pChild of pList.qa(".li_active").entries()){
			pChild[1].classList.remove("li_active")
		}

		// Get Parents
		var pParent = pWrapper
		while (pParent !== null){
			if (pParent.tagName == "LI") {
				pParent.classList.add("li_selected")
			}

			if (pParent.classList.contains('RootCollection')){break}

			pParent = pParent.parentElement
		}
		pWrapper.parentElement.classList.add("li_active")

		GenerateParams(GetElementParams(pLI))
		GenerateCollection(GetElementCollections(pLI))
	}
})
//=====================================================================================================
// Parameter Input
//=====================================================================================================
customElements.define('param-input',
class extends HTMLElement {
	constructor() {
		super();
	}

	def(sName, sType, pParam, sValueLabel){
		tInput = document.createElement("input");

		var tThis = this
		tThis.Name = sName
		tThis.Type = sType
		tThis.XML_Node = pParam

		var tInput = null
		switch(sType){
		//-----------------------------------------------------------------
		// BOOLS
		//-----------------------------------------------------------------			
			case 'BoolValue':

				tThis.Value = tThis.XML_Node.q("m_bValue").innerHTML == 'true'

				tInput = document.createElement("input");
				tInput.type = 'checkbox'
				tInput.checked = tThis.Value
				break
			case 'SimpleBoolValue':

				tThis.Value = tThis.XML_Node.innerHTML == "true"

				tInput = document.createElement("input");
				tInput.type = 'checkbox'
				tInput.checked = tThis.Value
				break	
		//-----------------------------------------------------------------
		// STRINGS
		//-----------------------------------------------------------------			
			case 'StringValue':

				tThis.Value = tThis.XML_Node.q("m_Value").getAttribute("text")

				tInput = document.createElement("input");
				tInput.type = 'text'
				tInput.value = tThis.Value
				break
			case 'StringValue_Attr':

				tThis.Value = tThis.XML_Node.getAttribute("text")

				tInput = document.createElement("input");
				tInput.type = 'text'
				tInput.value = tThis.Value
				break	
		//-----------------------------------------------------------------
		// NUMBERS
		//-----------------------------------------------------------------	
			case 'IntValue':
			case 'FloatValue':

				let tValue = tThis.XML_Node.q("m_fValue") || tThis.XML_Node.q("m_nValue")
				tThis.Value = tValue.innerHTML
				if (sType === 'IntValue'){tThis.Value = parseInt(tThis.Value)} else {tThis.Value = parseFloat(tThis.Value)}

				tInput = document.createElement("input");
				tInput.type = 'number'
				tInput.value = tThis.Value
				break	
		//-----------------------------------------------------------------
		// REFERENCES
		//-----------------------------------------------------------------	
			case 'ArtDefReferenceValue':
			case 'BLPEntryValue':

				let pName = tThis.XML_Node.q(sValueLabel || "NULL") || tThis.XML_Node.q("m_ElementName") || tThis.XML_Node.q("m_EntryName")
				tThis.ValueLabel = sValueLabel
				tThis.Value = pName.getAttribute("text")
				tThis.MainName = pName.tagName

				tInput = document.createElement("input");
				tInput.type = 'button'
				tInput.value = tThis.Value
				tInput.Type = "RefEditor"
				tInput.DataType = sType

				tInput.onclick = function(){
					GenerateReferences(tThis, tThis.XML_Node)
				}
				break				
		//-----------------------------------------------------------------
		// RGB
		//-----------------------------------------------------------------									
			case 'RGBValue':

				let R = parseInt(tThis.XML_Node.q("m_r").innerHTML)
				let G = parseInt(tThis.XML_Node.q("m_g").innerHTML)
				let B = parseInt(tThis.XML_Node.q("m_b").innerHTML)
				tThis.Value = rgbToHex(R, G, B).toUpperCase()

				tInput = document.createElement("input");
				tInput.type = 'text'
				tInput.value = tThis.Value
				$(tInput).colorPicker({
					renderCallback: function($elm, toggled) {
						tThis.onValueChange(tThis, tInput)
    				}
				})
				break
		}
		//-----------------------------------------------------------------
		//-----------------------------------------------------------------
		if(tInput){
			tInput.XML_Node = tThis.XML_Node

			var jInput = $(tInput)

			jInput.on("change paste keyup", function(){
				tThis.onValueChange(tThis, tInput)
			})

			tThis.appendChild(tInput)
		}
		return tThis
	}

	onValueChange(tThis, tInput){
		var sType = tThis.Type

		switch(sType){
		//-----------------------------------------------------------------
		// BOOLS
		//-----------------------------------------------------------------			
			case 'BoolValue':
				tThis.Value = tInput.checked
				tThis.XML_Node.q("m_bValue").innerHTML = tInput.checked

				//console.log(tThis.Value)
				break
			case 'SimpleBoolValue':
				tThis.Value = tInput.checked
				tThis.XML_Node.innerHTML = tInput.checked

				//console.log(tThis.Value)
				break	
		//-----------------------------------------------------------------
		// STRINGS
		//-----------------------------------------------------------------			
			case 'StringValue':
				tThis.Value = tInput.value
				tThis.XML_Node.q("m_Value").setAttribute("text", tInput.value)

				//console.log(tThis.Value)
				break
			case 'StringValue_Attr':
				tThis.Value = tInput.value
				tThis.XML_Node.setAttribute("text", tInput.value)

				let pElement = tThis.XML_Node.Refresh
				if (pElement !== null){
					if(pElement){
						let sTag = pElement.tagName
						if (sTag == "SIDEBARLIST-LI"){
							pElement.q(".li_wrapper > .li_label").innerHTML = tThis.Value
						}
						if (sTag == "PARAM-INPUT"){
							pElement.q("input").value = tThis.Value
						}
					}				
				}

				//console.log(tThis.Value)
				break	
		//-----------------------------------------------------------------
		// NUMBERS
		//-----------------------------------------------------------------	
			case 'IntValue':
			case 'FloatValue':
				let iValue = tInput.value
				if (sType === 'IntValue'){iValue = parseInt(iValue)} else {iValue = parseFloat(iValue)}

				let tValue = tThis.XML_Node.q("m_fValue") || tThis.XML_Node.q("m_nValue")
				tInput.value = iValue
				tThis.Value = iValue
				tValue.innerHTML = iValue
				
				//console.log(tThis.Value)
				break	
		//-----------------------------------------------------------------
		// REFERENCES
		//-----------------------------------------------------------------	
			case 'ArtDefReferenceValue':
			case 'BLPEntryValue':
				break				
		//-----------------------------------------------------------------
		// RGB
		//-----------------------------------------------------------------									
			case 'RGBValue':
				tThis.Value = tInput.value

				let tColor = hexToRgb(tThis.Value)

				tThis.XML_Node.q("m_r").innerHTML = tColor.r
				tThis.XML_Node.q("m_g").innerHTML = tColor.g
				tThis.XML_Node.q("m_b").innerHTML = tColor.b

				//console.log(tColor)
				break
		}
		//-----------------------------------------------------------------
		//-----------------------------------------------------------------		
	}

	pasteArtdef(pXML_Node){
		this.XML_Node.innerHTML = pXML_Node.innerHTML

		let pName = this.XML_Node.q(this.ValueLabel || "NULL") || this.XML_Node.q("m_ElementName") || this.XML_Node.q("m_EntryName")
		this.Value = pName.getAttribute("text")
		this.MainName = pName.tagName

		let tInput = this.firstChild
		tInput.value = this.Value
		tInput.pXML_Node = this.XML_Node

		tInput.onclick = () => {
			GenerateReferences(this, this.XML_Node)
		}
	}
})
//=====================================================================================================
// Parameters Table
//=====================================================================================================
customElements.define('param-table',
class extends HTMLElement {
	constructor() {
		super();
		this.Table		= null
		this.sName		= null
		this.Type		= null
		this.Value		= null
		this.XML_Node	= null
	}

	addParam(sName, sType, pParam){
		if (!this.Table){this.Table = document.createElement("Table"); this.appendChild(this.Table)}

		var tRow = document.createElement("tr"); //tRow.setAttribute('title', this.Type)

		tRow.Name		= sName
		tRow.Type		= sType
		tRow.XML_Node	= pParam

		var tLabel = document.createElement("td"); tLabel.innerHTML = tRow.Name
		var tControl = document.createElement("td")

		var tInput = document.createElement("param-input").def(sName, sType, pParam)
		tControl.appendChild(tInput)

		tRow.appendChild(tLabel); tRow.appendChild(tControl)
		this.Table.appendChild(tRow)

		return tRow
	}
})
//=====================================================================================================
// Collections Table
//=====================================================================================================
customElements.define('collec-table',
class extends HTMLElement {
	constructor() {
		super();
		this.Header		= null
		this.Template	= []
	}

	defineTypes(tTypes, pXML_Node, bIsCV){
		if (!this.Table){this.Table = document.createElement("Table"); this.appendChild(this.Table)}
		
		// We wanna start from scratch if the table defenition changes
		this.Table.innerHTML = ""
		this.Header = null; this.Template = []
		this.XML_Node = pXML_Node; this.CV = bIsCV

		var tRow = document.createElement("tr");

		for (let tType of tTypes.values()){
			let tHead = document.createElement("th");
			tHead.innerHTML		= tType[0]
			tHead.Name			= tType[0]
			tHead.Type			= tType[1]

			tRow.appendChild(tHead)
			this.Template.push([tHead.Name, tHead.Type])
		}
		let tHead = document.createElement("th");
		tRow.appendChild(tHead)

		this.Header = tRow
		this.Table.appendChild(tRow)
		this.newRowButton()
	}

	addRow(tRowData, pXML_Node, sValueLabel){
		var tRow = document.createElement("tr"); tRow.XML_Node = pXML_Node;
		for (let tType of this.Template.values()){

			let sName = tType[0]
			let sType = tType[1]
			let pParam = tRowData.get(sName)

			let tControl = document.createElement("td")
			let tInput = document.createElement("param-input").def(sName, sType, pParam, sValueLabel)
			tControl.appendChild(tInput); tRow.appendChild(tControl)
		}
		this.Table.insertBefore(tRow, this.NewRow)
		this.RemoveRowButton(tRow)
		return tRow			
	}

	newRowButton(){
		this.NewRow = document.createElement("tr");
		let tControl = document.createElement("td"); tControl.setAttribute('colspan', this.Template.length);
		let tInput = document.createElement("button"); tInput.classList.add('new-row-button')
		tInput.type = 'button'
		tInput.innerHTML = "<div></div>"

		tInput.onclick = () => {
			this.newRow()
		}

		tControl.appendChild(tInput); this.NewRow.appendChild(tControl); this.Table.appendChild(this.NewRow)
	}

	RemoveRowButton(tRow){
		let tControl = document.createElement("td");
		let tInput = document.createElement("button"); tInput.classList.add('remove-row-button')
		tInput.type = 'button'
		tInput.innerHTML = "<div></div>"

		tInput.onclick = () => {
			this.removeRow(tRow)
		}

		tControl.appendChild(tInput); tRow.appendChild(tControl);
	}

	removeRow(tRow){
		let pXML_Node = tRow.XML_Node
		pXML_Node.remove()
		tRow.remove()

		// If there are only TRs then the only rows remaining are the header and remove button
		if (this.Table.qa("tr").length <= 2){
			document.getElementById("collec_table_slot").innerHTML = ""
			document.getElementById("collec_table_slot").style.display = 'none'
		}
	}

	newRow(){
		if (this.CV){
			let iNewKey = this.XML_Node.qa("[class$=\"CollectionValue\"] > m_Values > Element").length
			let pValues = this.XML_Node.q("[class$=\"CollectionValue\"] > m_Values")
			let pXML_Node = pValues.qa("Element")[iNewKey -1].cloneNode(true); pValues.appendChild(pXML_Node)
			
			pXML_Node = pValues.qa("Element")[iNewKey]
			this.generateRows([pXML_Node])
		} else {
			let iNewKey = this.XML_Node.qa(":scope > Element").length
			let pValues = this.XML_Node
			let pTemplate = pValues.qa(":scope > Element")[iNewKey -1]
			let pXML_Node = pTemplate.cloneNode(true); pValues.appendChild(pXML_Node)

			pXML_Node = pValues.qa(":scope > Element")[iNewKey]
			this.generateRows([pXML_Node])			
		}
		//console.log(this.XML_Node)
	}

	generateRows(tChildren){
		//-----------------------------------------------------------------
		// CollectionValues
		//-----------------------------------------------------------------		
		if (this.CV){
			for (let tChild of tChildren.values()) {
				//console.log(tChild)
				let tParams = new Map()

				//tParams.set('m_ParamName', tChild.q("m_ParamName"))
				tParams.set(this.Template[0][0], tChild)
				this.addRow(tParams, tChild, "m_ParamName")
			}
		//-----------------------------------------------------------------
		// ChildCollection
		//-----------------------------------------------------------------		
		} else {

			for (let tChild of tChildren.values()) {
				let tParams = new Map()
				tParams.set('m_Name', tChild.q("m_Name"))

				//console.log(tChild)
				//----------------------------------------------
				// Individual Values in the Child Element
				//----------------------------------------------
				let tChildParams = tChild.qa(":scope > m_Fields > m_Values > Element")
				for (let tChildParam of tChildParams.values()){
					let sName = tChildParam.q("m_ParamName").getAttribute("text")
					tParams.set(sName, tChildParam)
				}
				//----------------------------------------------
				//----------------------------------------------

				this.addRow(tParams, tChild)
			}
		}		
	}
})
//=====================================================================================================
// Context Menu
//=====================================================================================================
// Globals
//-----------------------------------------------------------------
let g_RightClickPosition = null
//-----------------------------------------------------------------
// Common Menu Items
//-----------------------------------------------------------------
const mi_InspectElement = new MenuItem({
	label: 'Inspect Element',
	id: 'inspectElement',
	click: () => {
		remote.getCurrentWindow().inspectElement(g_RightClickPosition.x, g_RightClickPosition.y)
	}
})

const menuDefault = new Menu()
menuDefault.append(mi_InspectElement)
//-----------------------------------------------------------------
// li_ContextMenu Items
//-----------------------------------------------------------------
const mi_RemoveLI = new MenuItem({
	label: 'Remove',
	id: 'removeLI',
	click: () => {
		let pTarget = menuLI.Target
		let pXML_Node = pTarget.XML_Node

		pTarget.remove()
		pXML_Node.remove()
	}
})

const mi_DuplicateLI = new MenuItem({
	label: 'Duplicate',
	id: 'duplicateLI',
	click: () => {
		let pTarget = menuLI.Target.parentElement
		let pXML_Temp = menuLI.Target.XML_Node
		let pXML_Node = pXML_Temp.cloneNode(true)

		while (pTarget.tagName !== "SIDEBARLIST-LI"){
			pTarget = pTarget.parentElement
		}
		AppendChild(pTarget, pXML_Node)
		pXML_Temp.parentElement.appendChild(pXML_Node)
	}
})

const mi_CopyLI = new MenuItem({
	label: 'Copy',
	id: 'copyLI',
	click: () => {
		let pXML_Temp = menuLI.Target.XML_Node.cloneNode(true)
		clipboard.writeText(pXML_Temp.outerHTML)
	}
})

const mi_PasteLI = new MenuItem({
	label: 'Paste',
	id: 'pasteLI',
	click: () => {
		let pTarget = menuLI.Target
		let pTargetXML = menuLI.Target.XML_Node

		let pXML_Node = menuLI.XML_Node.cloneNode(true)

		AppendChild(pTarget, pXML_Node)
		pTargetXML.appendChild(pXML_Node)
	}
})

const menuLI = new Menu()
menuLI.append(mi_InspectElement)
menuLI.append(new MenuItem({ type: 'separator' }))
menuLI.append(mi_CopyLI)
menuLI.append(mi_PasteLI)
menuLI.append(new MenuItem({ type: 'separator' }))
menuLI.append(mi_RemoveLI)
menuLI.append(mi_DuplicateLI)
//-----------------------------------------------------------------
// Artdef Input Items
//-----------------------------------------------------------------
const mi_CopyArtdef = new MenuItem({
	label: 'Copy',
	id: 'copyArtdef',
	click: () => {
		let pXML_Node = menuArtdef.Target.XML_Node
		clipboard.writeText(pXML_Node.outerHTML)
	}
})

const mi_PasteArtdef = new MenuItem({
	label: 'Paste',
	id: 'pasteArtdef',
	click: () => {
		let pParamInput = menuArtdef.Target.parentElement
		pParamInput.pasteArtdef(menuArtdef.XML_Node)
	}
})

const menuArtdef = new Menu()
menuArtdef.append(mi_InspectElement)
menuArtdef.append(new MenuItem({ type: 'separator' }))
menuArtdef.append(mi_CopyArtdef)
menuArtdef.append(mi_PasteArtdef)
//-----------------------------------------------------------------
// Context Menu Handler
//-----------------------------------------------------------------
window.addEventListener('contextmenu', (e) => {

	e.preventDefault()
	g_RightClickPosition = {x: e.x, y: e.y}

	let pTarget = e.target
	//-----------------------------------------------------------------
	// Sidebar
	//-----------------------------------------------------------------
	if (pTarget.classList.contains('sidebarlist_li')){
		//Disable Copy and Paste
		menuLI.getMenuItemById('copyLI').enabled = false
		menuLI.getMenuItemById('pasteLI').enabled = false

		// Enable duplicating and removing
		menuLI.getMenuItemById('duplicateLI').enabled = true
		menuLI.getMenuItemById('removeLI').enabled = true

		let pElement = pTarget
		while (pElement.tagName !== "LI"){
			pElement = pElement.parentElement
		}

		if (pElement.classList.contains('elementLI')){
			menuLI.getMenuItemById('copyLI').enabled = true
		} else {

			// You can't remove or duplicate non-elements
			menuLI.getMenuItemById('duplicateLI').enabled = false
			menuLI.getMenuItemById('removeLI').enabled = false

			let sClipboard = clipboard.readText()
			let pXML_Node = pXML.createElement('template'); pXML_Node.innerHTML = sClipboard
			pXML_Node = pXML_Node.firstChild

			if (pXML_Node && pXML_Node.querySelector && IsElement(pXML_Node)){
				menuLI.getMenuItemById('pasteLI').enabled = true
				menuLI.XML_Node = pXML_Node
			}
		}
		menuLI.Target = pElement.parentElement
		menuLI.popup(remote.getCurrentWindow())
		return			
	}
	//-----------------------------------------------------------------
	// Artdef/BLP ref
	//-----------------------------------------------------------------
	if (pTarget.tagName == "INPUT" && pTarget.Type == "RefEditor"){
		menuArtdef.Target = pTarget
		let sTargetType = pTarget.DataType
		menuArtdef.getMenuItemById('pasteArtdef').enabled = false

		let sClipboard = clipboard.readText()
		let pXML_Node = pXML.createElement('template'); pXML_Node.innerHTML = sClipboard; pXML_Node = pXML_Node.firstChild

		if (pXML_Node && pXML_Node.tagName == "Element" && pXML_Node.getAttribute("class") && pXML_Node.getAttribute("class").GetCoreClass() == sTargetType) {
			menuArtdef.XML_Node = pXML_Node
			menuArtdef.getMenuItemById('pasteArtdef').enabled = true
		}

		menuArtdef.popup(remote.getCurrentWindow())
		return			
	}
	menuDefault.popup(remote.getCurrentWindow())
}, false)